import java.io.File;
import java.io.IOException;
import javax.sound.sampled.*;
import javax.sound.sampled.Line.Info;

import java.io.ByteArrayOutputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;

// TODO: ミュート解除せねば
// TODO: ミキサーのマスターボリュームを変化させたい。Win10 では API 叩かなければならないのか。

/*
 * とにかく Sound システムはややこしい。
 * - OSのマスターボリュームは、Java（に限らず、アプリ）から操作するのは難しい。
 * - 入力/出力に source/target との用語を使う。
 * - データ構造が再帰的なので、全列挙には再起呼び出しが必要。
 * - Mixer 自身が Line でもあって、入出力以外にも、
 *   自分自身が Control を持てるデータ構造になっている。（が、持たない。）
 * - Mixer は Line のサブクラスだが、Mixer.Info は Line.Info のサブクラスではない。
 * - Mixer や Line は、直接は（ほぼ）入手できなくて、Info を鍵にして探し出す。
 * - その Info も、自分で作り出すのはほぼ無理で、Mixer/Line にぶら下がっているものを入手する。
 * - Mixer や Line には Control がぶら下がっている。
 * - Windows では、再生に使う Mixer と、ボリューム調整に使う Mixer が異なるが、
 *   2つの Mixer の関係性が、name が似ている以外になさそう。
 *   しかもその名前が、バッファをあふれて途中で切れている。（爆）
 *   Linux だと１つの Mixer でよさそう。
 */
public class Audio extends Thread {
    public static int SAMPLING_RATE = 48000;
    public static int SOUND_FREQ = 880;
    /** サイン波を再生する回数（0ならファイルを再生する）*/
    private int times;
    /** サイン波の長さ */
    private int msec;
    /** 再生する音声ファイル */
    private String file;
    /** 再生に用いるデバイス */
    private String targetName;
    /** 音量調整に用いるデバイス */
    private String mixerName;
    /** 音量 (0--100) */
    private int volume = 50;
    /** 再生する音声ストリーム（内部形式）*/
    private AudioInputStream audio;
    
    public Audio(String targetName, String mixerName, int volume, int times, int msec) {
    	this.targetName = targetName;
    	this.mixerName  = mixerName;
    	this.volume = volume;
        this.times = times;
        this.msec  = msec;
        audio = new WaveBeep(SAMPLING_RATE, SOUND_FREQ, msec/2);
    }
    
    public Audio(String targetName, String mixerName, int vol, String file) {
    	this.targetName = targetName;
    	this.mixerName  = mixerName;
        this.volume = vol;
        this.file = file;
       
        try {
            audio = AudioSystem.getAudioInputStream(new File(file));
        } catch (UnsupportedAudioFileException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    public Mixer.Info getMixerInfoByName(String name) {
    	Mixer.Info[] info = AudioSystem.getMixerInfo();
    	for (int i=0; i<info.length; i++) {
            if (deMessyCode(info[i].getName()).equals(name)) return info[i];
    	}
    	return null;
    }
    
    /** Clip/Port の存在する MixerInfo のリストを、文字列展開して返す */
    public static String[][] getMixerNames(boolean isClip) {
        ArrayList<String[]> ret = new ArrayList<String[]>();
        for (Mixer.Info mixerInfo: AudioSystem.getMixerInfo()) {
            String name = deMessyCode(mixerInfo.getName());
            String desc = deMessyCode(mixerInfo.getDescription());
            String vend = deMessyCode(mixerInfo.getVendor());
            String ver  = deMessyCode(mixerInfo.getVersion());
            Mixer mixer = AudioSystem.getMixer(mixerInfo);
            boolean exist = false;
            if (isClip) { // (Clip)
                for (Line.Info lineInfo: mixer.getSourceLineInfo()) {
                    try {
                        if (mixer.getLine(lineInfo) instanceof Clip) exist = true;
                    } catch (LineUnavailableException e) {
                        //e.printStackTrace();
                    }
                }
            } else { // Target (Port)
                for (Line.Info lineInfo: mixer.getTargetLineInfo()) {
                    try {
                        if (mixer.getLine(lineInfo) instanceof Port) exist = true;
                    } catch (LineUnavailableException e) {
                        //e.printStackTrace();
                    }
                }
            }
            if (exist) ret.add(new String[] {name, desc, vend, ver});
        }
    	return ret.toArray(new String[ret.size()][4]);
    }
    
    @Override
    public void run() {
    	Mixer.Info mixerInfo = null; // volume 調整用
    	if (!mixerName.isEmpty()) {
            mixerInfo = getMixerInfoByName(mixerName);
            if (mixerInfo == null) {
                System.out.println("No device for mixer " + mixerName);
                //return;  // continue
            }
    	}
    	Mixer.Info targetInfo = null; // 再生用
    	if (!targetName.isEmpty()) {
            targetInfo = getMixerInfoByName(targetName);
            if (targetInfo == null) {
                System.out.println("No device for target " + targetName);
                return;
            }
    	}
        
        //System.out.println("Sound Start");
    	try {
            Mixer mixer = AudioSystem.getMixer(mixerInfo);
            for (Control c: getSpecifiedControls(mixer, FloatControl.Type.VOLUME)) {
            	FloatControl fc = (FloatControl)c;
            	System.out.println("setVolume : " +fc);
            	fc.setValue(fc.getMinimum()+(volume/100f)*(fc.getMaximum()-fc.getMinimum()));
            }
            for (Control c: getSpecifiedControls(mixer, BooleanControl.Type.MUTE)) {
            	BooleanControl bc = (BooleanControl)c;
            	System.out.println("unsetMute : " +bc);
            	bc.setValue(false);
            }
            
            Clip clipTarget = AudioSystem.getClip(targetInfo);
            clipTarget.open(audio);
            if (times > 0) {
            	System.out.println("sin wave " + msec + "(ms) * " + times);
            	for (int i=0; i<times; i++) {
                    clipTarget.setFramePosition(0);
                    clipTarget.start();
                    sleep(msec);
            	}
            } else {
            	System.out.println("Sound file \"" + file + "\"");
                clipTarget.start();
                sleep(clipTarget.getMicrosecondLength()/1000);
            }
            clipTarget.close();
        } catch (LineUnavailableException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        } catch (NegativeArraySizeException e) {
            //e.printStackTrace(); // PulseAudioClip.java:450
        }
        //System.out.println("Sound End");
    }
    
    private static ArrayList<Control> getSpecifiedControls(Mixer mixer, Control.Type type) {
    	System.out.println("setSpecifiedFloatControl: mixer = " +mixer);
    	ArrayList<Control> ret = new ArrayList<Control>();
    	ArrayList<Line.Info> a = new ArrayList<Line.Info>();
    	a.addAll(Arrays.asList(mixer.getTargetLineInfo()));
        //a.addAll(Arrays.asList(mixer.getSourceLineInfo()));
    	for (Line.Info lineInfo: a) {
            try {
                Line lineOut = mixer.getLine(lineInfo);
                System.out.println("search mixer = "
                                   + deMessyCode(mixer.getMixerInfo().getName())
                                   + "; line = "+ deMessyCode(lineInfo.toString()));
                lineOut.open();
                //Control c = lineOut.getControl(type); // CompoundControlの中までは探してくれない 
                getSpecifiedControls(ret, lineOut.getControls(), type);
                //System.out.println(lineOut.isControlSupported(type) + " vs. "+ ret.isEmpty());
                lineOut.close();
            } catch (Exception e) {
                //e.printStackTrace();
            }
    	}
    	return ret;
    }
    private static void getSpecifiedControls(ArrayList<Control> ctrls, Control[] controls, Control.Type type) {
    	for (Control c: controls) {
            if (c instanceof CompoundControl) { // 再帰的に探す
                getSpecifiedControls(ctrls, ((CompoundControl) c).getMemberControls(), type);
            } else {
                if (c.getType() == type) {
                	ctrls.add(c);
                    System.out.println(" found ctrl = " + c);
                }
            }
    	}
    }
    
    /**
     * JDKのバグ対策（サウンドデバイス名のエンコード間違いを直す）
     * https://qiita.com/TakaW/items/023b06cd1ba18ca62047
     * https://bugs.java.com/bugdatabase/view_bug.do?bug_id=JDK-8177951
     * 
     *  Windows の eclipse で開発する場合は、-Dfile.encoding=MS932 を実行の構成の VM オプションに追加する。
     *  https://www.gwtcenter.com/eclipse-sets-file-encoding
     */
    private static final byte HIGH_BIT = (byte) 0b11000000;
    public static String deMessyCode(String messyCode) {
        ByteArrayOutputStream buf = new ByteArrayOutputStream(messyCode.length());
        byte[] originalBytes = messyCode.getBytes(Charset.forName("UTF-8"));
        String systemEnc = System.getProperty("file.encoding");
        //System.out.println(systemEnc + (systemEnc.equalsIgnoreCase("MS932")));
        
        for (int i = 0; i < originalBytes.length; i++) {
            if ((byte) (originalBytes[i] & HIGH_BIT) == HIGH_BIT) {
            	if (i==originalBytes.length-2 && systemEnc.equalsIgnoreCase("MS932")) break;
            	// SJIS の泣き別れが起きないよう、最後の1文字は書き込まない
                buf.write((originalBytes[i] << 6) | (originalBytes[i + 1] << 2 >>> 2));
                // DELETE 0b110000 and move 6, then delete the 0b10 prefix of the second byte.
                i++;
            } else {
                buf.write(originalBytes[i]);
            }
        }
        return new String(buf.toByteArray(), Charset.forName(systemEnc));
    }
}
